# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Milestone events are some type of events that would just occur at some point.
Nothing special about them, they just happen and something is instantly done.
"""

from typing_extensions import override

from titanq._event import Event
from titanq._event.sink import Sink


class PreparingDataEvent(Event):

    @override
    def output(self, sink: Sink) -> None:
        with sink.content_builder() as cb:
            cb.add_emoji(":wrench:")
            cb.add_whitespace()
            cb.add_text("Preparing data for TitanQ...")


class UploadStartEvent(Event):

    @override
    def output(self, sink: Sink) -> None:
        with sink.content_builder() as cb:
            cb.add_emoji(":outbox_tray:")
            cb.add_whitespace()
            cb.add_text("Starting data upload...")

class UploadEndEvent(Event):

    @override
    def output(self, sink: Sink) -> None:
        with sink.content_builder() as cb:
            cb.add_emoji(":clock1:")
            cb.add_whitespace()
            cb.add_text("Upload completed!")


class SendingRequestEvent(Event):

    @override
    def output(self, sink: Sink) -> None:
        with sink.content_builder() as cb:
            cb.add_emoji(":rocket:")
            cb.add_whitespace()
            cb.add_text("Sending request to the server...")


class OptimizationCompletedEvent(Event):

    @override
    def output(self, sink: Sink) -> None:
        with sink.content_builder() as cb:
            cb.add_emoji(":white_check_mark:")
            cb.add_whitespace()
            cb.add_text("Optimization completed!")


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing_extensions import override

from pyfiglet import Figlet
from rich.console import Console
from rich.panel import Panel

from titanq._event import Sink
from titanq._event.sink import ContentBuilder

# if set to true will add the path to the source where it was logged from.
# no useful in this case we do not manage the source and will only lead
# to this file.
_LOG_PATH = False

# if set to true, will repeat timestamps on sequent logs even if the same
_REPEAT_TIMESTAMPS = True

# This option disable the option that forces the logger to be optimized for jupyter.
# Disabling this avoid logs to be split into multiple cells
# which isn't in our case (we will need to use Live feature).
_FORCE_JUPYTER_RENDERING = False


class PrettyTextContentBuilder(ContentBuilder):

    def __init__(self, console: Console):
        self._console = console
        self._content = ""

    @override
    def add_emoji(self, emoji: str) -> None:
        self._content += emoji

    @override
    def add_text(self, text: str) -> None:
        self._content += text

    @override
    def add_whitespace(self) -> None:
        self._content += " "

    @override
    def end(self) -> None:
        self._console.log(self._content)


class PrettyTextSink(Sink):
    """
    The PrettyTextSink utilizes rich library to enhance the output and make it
    as beautiful as possible and pleasing for the end-users.

    It offers features such as Emoji's support, colors and some progress bars.
    """

    def __init__(self, banner_header_text: str):
        """
        Initialize the pretty sink logger.

        :param banner_header_text: Text added to the top of the banner
        """
        self._console = Console(log_path=_LOG_PATH, force_jupyter=_FORCE_JUPYTER_RENDERING)
        self._console._log_render.omit_repeated_times = not _REPEAT_TIMESTAMPS

        self._display_banner(header=banner_header_text)

    @override
    def content_builder(self) -> ContentBuilder:
        return PrettyTextContentBuilder(self._console)

    def _display_banner(self, header: str) -> None:
        """
        Displays a banner.
        """
        #╭───────────────── Version vX.Y.Z ──────────────────╮
        #│      _________________              _______       │
        #│      ___  __/__(_)_  /______ _________  __ \      │
        #│      __  /  __  /_  __/  __ `/_  __ \  / / /      │
        #│      _  /   _  / / /_ / /_/ /_  / / / /_/ /       │
        #│      /_/    /_/  \__/ \__,_/ /_/ /_/\___\_\       │
        #│                                                   │
        #│                                                   │
        #╰───────────────────────────────────────────────────╯
        panel = Panel.fit(
            Figlet(font='speed').renderText("TitanQ"), style="purple",
            title=f"[italic bold purple]{header}[/]",
            padding=(0, 6),  # (top_bottom, left_right)
            border_style="purple",
        )
        self._console.print(panel)


# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
from typing import List, Union

from titanq._event.sink import Sink
from titanq._event.sink.pretty_text import PrettyTextSink
from titanq._event.sink.text import TextSink
from titanq._version import __version__ as titanq_version


class Event(ABC):

    @abstractmethod
    def output(self, sink: Sink):
        """Output self to the sink."""


class EventEmitter():
    """
    A class responsible for emitting events to subscribed output sinks.
    """

    def __init__(self):
        self._sinks: List[Sink] = []

    def subscribe(self, sink: Sink) -> None:
        """subscribe a sink to the events."""
        self._sinks.append(sink)

    def emit(self, event: Event) -> None:
        """Emit an event to the subscribed sinks."""
        for sink in self._sinks:
            event.output(sink)


class EventEmitterFactory():
    """
    Creates an EventEmitter.
    None, on or multiple sinks will be subscribed based on the `log_mode`.
    """

    def __init__(self):
        self._event_emitter = EventEmitter()

    def create(self, log_mode: Union[str, List[str], None]) -> EventEmitter:
        """Setup all sinks based on the given mode and returns the EventEmitter"""

        if log_mode is None or log_mode == "off":
            return self._event_emitter

        elif isinstance(log_mode, str):
            self._event_emitter.subscribe(self._new_sink(log_mode))

        elif isinstance(log_mode, list):
            for mode in log_mode:
                self._event_emitter.subscribe(self._new_sink(mode))

        else:
            raise ValueError("log_mode must be a string, list of strings, or None")

        return self._event_emitter

    def _new_sink(self, type: str) -> Sink:
        if type == "pretty":
            return PrettyTextSink(titanq_version)
        elif type == "text":
            return TextSink(f"Using TitanQ SDK {titanq_version}")
        else:
            raise ValueError(f"got an unknown log_mode type: {type}")


# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod


class ContentBuilder(ABC):
    """
    This class acts as a context manager to ensure that any necessary
    output build logic is properly executed at the end of content building.

    Usage example:
        with ConcreteContentBuilder() as cb:
                cb.add_emoji("👋")
                cb.add_text("Hello")
    """

    def __enter__(self) -> 'ContentBuilder':
        return self

    def __exit__(self, ex_type, ex_value, ex_trace_back):
        self.end()

    @abstractmethod
    def add_emoji(self, emoji: str) -> None:
        """Add an emoji to the content."""

    @abstractmethod
    def add_text(self, text: str) -> None:
        """Add a text to the content."""

    @abstractmethod
    def add_whitespace(self) -> None:
        """Add a whitespace to the content"""

    @abstractmethod
    def end(self) -> None:
        """Finalize the content building process."""


class Sink(ABC):
    """
    Abstract class of a Sink type object. A Sink is any output type that you could
    use to transfer information from somewhere to another destination.

    As an example, a logger is a Sink.
    """

    @abstractmethod
    def content_builder(self) -> ContentBuilder:
        """Returns a content builder to be able to build an output."""


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing_extensions import override

import logging

from titanq._event.sink import ContentBuilder, Sink


# Configure the logger
logging.basicConfig(
    format='[%(asctime)s] %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.INFO
)


class TextContentBuilder(ContentBuilder):

    def __init__(self):
        self._content = ""

    @override
    def add_emoji(self, emoji: str) -> None:
        # no emoji for text based output
        pass

    @override
    def add_text(self, text: str) -> None:
        self._content += text

    @override
    def add_whitespace(self) -> None:
        self._content += " "

    @override
    def end(self) -> None:
        logging.info(self._content.strip())


class TextSink(Sink):
    """The TextSink is just a simple python logger."""

    def __init__(self, header_text: str):
        """
        Initialize a simple text sink logger.

        :param header: Text logged when initialising the logger
        """
        logging.info(header_text)

    @override
    def content_builder(self) -> ContentBuilder:
        return TextContentBuilder()